;; Generated package description from cursory.el  -*- no-byte-compile: t -*-
(define-package "cursory" "1.0.1.0.20240913.70428" "Manage cursor styles using presets" '((emacs "27.1")) :commit "d079b397ba7270336a74727c990c4b3992846cfe" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("convenience" "cursor") :url "https://github.com/protesilaos/cursory")
